import numpy as np
import re

from numpy.core.fromnumeric import argmax


def get_Several_MinMax_Array(np_arr, several):
    """
    获取numpy数值中最大或最小的几个数
    :param np_arr:  numpy数组
    :param several: 最大或最小的个数（负数代表求最大, 正数代表求最小）
    :return:
        several_min_or_max: 结果数组
    """
    np_arr = np.array(np_arr)
    if several > 0:
        if several == len(np_arr):
            index_pos = np.argmin(np_arr)
        else:
            index_pos = np.argpartition(np_arr, several)[:several]
    else:
        if several == len(np_arr):
            index_pos = np.argmax(np_arr)
        else:
            index_pos = np.argpartition(np_arr, several)[several:]
    several_min_or_max = np_arr[index_pos]
    return index_pos, several_min_or_max


def get_non_zero_parts(signal):
    signal = np.array(signal) != 0
    assert signal.ndim == 1
    if np.max(signal) == 0:
        return [], []
    temp_signal = np.hstack([0, signal, 0])
    diff_signal = temp_signal[1:] - temp_signal[:-1]
    starts = np.nonzero(diff_signal > 0)[0]
    ends = np.nonzero(diff_signal < 0)[0]
    assert len(starts) == len(ends)
    return starts, ends


def find_longest_start_end(arr):
    substr = max(re.findall('1+', str(arr)))
    obj = re.search(substr, str(arr))
    return obj.start(), obj.end()


def get_longest_part(signal):
    if len(signal) == 0 or np.max(np.abs(signal)) == 0:
        return 0, 0

    signal = (signal != 0).astype('int8')
    str_signal = str(signal.tolist()).replace(', ', '')[1:-1]
    start, end = find_longest_start_end(str_signal)

    return start, end


def bilinear_interpolate(im, x, y):
    x = np.asarray(x)
    y = np.asarray(y)

    x0 = np.floor(x).astype(int)
    x1 = x0 + 1
    y0 = np.floor(y).astype(int)
    y1 = y0 + 1

    x0 = np.clip(x0, 0, im.shape[1]-1)
    x1 = np.clip(x1, 0, im.shape[1]-1)
    y0 = np.clip(y0, 0, im.shape[0]-1)
    y1 = np.clip(y1, 0, im.shape[0]-1)

    Ia = im[ y0, x0 ]
    Ib = im[ y1, x0 ]
    Ic = im[ y0, x1 ]
    Id = im[ y1, x1 ]

    wa = (x1-x) * (y1-y)
    wb = (x1-x) * (y-y0)
    wc = (x-x0) * (y1-y)
    wd = (x-x0) * (y-y0)

    return wa*Ia + wb*Ib + wc*Ic + wd*Id


def bilinear_by_meshgrid(image, x_grid, y_grid):

    #               Ia, Wd                          Ic, Wb
    #           (floor_x, floor_y)              (ceil_x, floor_y)   
    #
    #                               (x, y)
    #
    #               Ib , Wc                         Id, Wa
    #           (floor_x, ceil_y)               (ceil_x, ceil_y)   
    #

    assert image.shape == x_grid.shape == y_grid.shape
    assert image.ndim == 2
    H, W = image.shape[:2]

    floor_x_grid = np.floor(x_grid).astype('int32')
    floor_y_grid = np.floor(y_grid).astype('int32')
    ceil_x_grid = floor_x_grid + 1
    ceil_y_grid = floor_y_grid + 1

    if np.max(ceil_x_grid) > W -1 or  np.max(ceil_y_grid) > H -1 or np.min(floor_x_grid) < 0 or np.min(floor_y_grid) < 0:
        print("Warning: index value out of original matrix, a crop operation will be applied.")

        floor_x_grid = np.clip(floor_x_grid, 0, W-1).astype('int32')
        ceil_x_grid = np.clip(ceil_x_grid, 0, W-1).astype('int32')
        floor_y_grid = np.clip(floor_y_grid, 0, H-1).astype('int32')
        ceil_y_grid = np.clip(ceil_y_grid, 0, H-1).astype('int32')

    Ia = image[ floor_y_grid, floor_x_grid ]
    Ib = image[ ceil_y_grid, floor_x_grid ]
    Ic = image[ floor_y_grid, ceil_x_grid ]
    Id = image[ ceil_y_grid, ceil_x_grid ]

    wa = (ceil_x_grid - x_grid) * (ceil_y_grid - y_grid)
    wb = (ceil_x_grid - x_grid) * (y_grid - floor_y_grid)
    wc = (x_grid - floor_x_grid) * (ceil_y_grid - y_grid)
    wd = (x_grid - floor_x_grid) * (y_grid - floor_y_grid)

    assert np.min(wa) >=0 and np.min(wb) >=0 and np.min(wc) >=0 and np.min(wd) >=0
    
    W = wa + wb + wc + wd
    assert np.sum(W[:, -1]) + np.sum(W[-1, :]) == 0
    
    wa[:-1, -1] = ceil_y_grid[:-1, -1] - y_grid[:-1, -1]
    wb[:-1, -1] = y_grid[:-1, -1] - floor_y_grid[:-1, -1]
    
    wb[-1, :-1] = ceil_x_grid[-1, :-1] - x_grid[-1, :-1]
    wd[-1, :-1] = x_grid[-1, :-1] - floor_x_grid[-1, :-1]
    
    wd[-1, -1] = 1
    
    W = wa + wb + wc + wd
    assert np.max(W) == np.min(W) == 1
    
    res_image = wa*Ia + wb*Ib + wc*Ic + wd*Id

    return res_image


def get_start_and_end(data):
    data = np.array(data)
    assert data.ndim == 1
    signal = data != 0
    length = len(signal)
    start = np.argmax(signal)
    end = length - np.argmax(signal[::-1]) - 1
    return start, end



import numpy as np
import pandas as pd
from tqdm import tqdm

from pathlib import Path

from ..utils import ClassifierEvalMultilabel

from ..utils import get_TFPN
from .base import EvaluatorBase

from ..utils import cv_rgb_imread
from ..utils import cv_rgb_imwrite
from ..utils import encode_path
from ..utils import dir_check
from ..utils import puzzle


class ClassificationMultiLabelEvaluator(EvaluatorBase):
    TYPE = 'MultiLabel'

    def _get_result_dict(self, dm):
        result_dict = dict()
        for rec in dm:
            scores = dm.get_multilabel_scores_from_record(rec)
            result_dict[rec['info']['uuid']] = scores
        return result_dict

    def _get_gt_result_dict(self, dm_gt):
        return self._get_result_dict(dm_gt)

    def _get_pred_result_dict(self, dm_pred):
        return self._get_result_dict(dm_pred)

    def compute_wmAP(self, df):
        """ Compute mAP weighted by the actual number ground truth class """
        class_occurrence_list = np.sum(self.gt_data_list, axis=0).tolist()
        total_gt = max(1, np.sum(class_occurrence_list))
        class_weights = {key:val / total_gt for key, val in zip(self.class_list, class_occurrence_list)}
        wmAP = sum([class_weights[row.name] * row.AP for _, row in df.iterrows()])
        return wmAP

    def eval_miss_fa(self, threshold, class_list=None):
        
        annotation_distribution_list, prediction_distribution_list = self.gt_data_list, self.pred_data_list
        classnames = self.class_list
        if class_list is None:
            class_list = self.class_list

        for class_name in class_list:
            assert class_name in classnames, f"classname {class_name} not in classnames {classnames}."
        index_list = list(map(self._dm_gt.get_classid_from_classname, class_list))

        interested_ann_array = np.array(annotation_distribution_list)[:, index_list]
        interested_pred_array = (np.array(prediction_distribution_list)[:, index_list] > threshold).astype('float64')
        
        class_results = dict()
        
        for index in range(len(class_list)):
            ann_array = interested_ann_array[:, index]
            pred_array = interested_pred_array[:, index]
            info_dict = get_TFPN(ann_array, pred_array)

            class_results[class_list[index]] = info_dict
        return class_results

    def eval_ap(self, verbose=True):
        annotation_distribution_list = self.gt_data_list
        prediction_distribution_list = self.pred_data_list

        # compute ap
        inflection_info = ClassifierEvalMultilabel.get_inflection_info(
            y_true  = np.array(annotation_distribution_list),
            y_score = np.array(prediction_distribution_list),
            class_name_list = self.class_list
        )
        p_at_r_100 = ClassifierEvalMultilabel.compute_p_at_r(            
            y_true  = np.array(annotation_distribution_list),
            y_score = np.array(prediction_distribution_list),
            class_name_list = self.class_list,
            recall_thresh=1
        )
        thr_at_p_99 = ClassifierEvalMultilabel.get_threshold_at_p(            
            y_true  = np.array(annotation_distribution_list),
            y_score = np.array(prediction_distribution_list),
            class_name_list = self.class_list,
            precision_thresh=0.99
        )
        p_at_r_99 = ClassifierEvalMultilabel.compute_p_at_r(            
            y_true  = np.array(annotation_distribution_list),
            y_score = np.array(prediction_distribution_list),
            class_name_list = self.class_list,
            recall_thresh=0.99
        )
        p_at_r_95 = ClassifierEvalMultilabel.compute_p_at_r(            
            y_true  = np.array(annotation_distribution_list),
            y_score = np.array(prediction_distribution_list),
            class_name_list = self.class_list,
            recall_thresh=0.95
        )
        ap_dict = ClassifierEvalMultilabel.compute_ap(
            y_true  = np.array(annotation_distribution_list),
            y_score = np.array(prediction_distribution_list),
            class_name_list = self.class_list
        )

        ap_info = dict()
        for key in self.class_list:
            res_dict = {
                'ap': ap_dict[key],
                'inflection': inflection_info[key],
                'p_at_r_99': p_at_r_99[key],
                'p_at_r_100': p_at_r_100[key],
                'p_at_r_95': p_at_r_95[key],
                'thr_at_p_99': thr_at_p_99[key]
            }
            ap_info[key] = res_dict

        df = pd.DataFrame.from_dict(ap_dict, orient='index', columns=['AP'])
        df_notnull = df[df['AP'].notnull()]  # remove NaN data (No samples of this class in gt)
        wmAP = self.compute_wmAP(df_notnull)

        # print evaluation
        if verbose:
            if len(df_notnull) > 1:
                # inflection info
                print()
                print('---------------')
                print('inflection info')
                df_inflection = pd.DataFrame.from_dict(inflection_info).T
                print(df_inflection)

                # p_at_r_99
                print()
                print('---------------')
                print('precision at recall 0.99')
                p_at_r_99_pd = pd.DataFrame.from_dict(p_at_r_99, orient='index', columns=['precision_at_recall_0.99'])
                print(p_at_r_99_pd)

                # threshold at precision 0.99
                print()
                print('---------------')
                print('threshold at precision 0.99')
                thr_at_p_99_pd = pd.DataFrame.from_dict(thr_at_p_99, orient='index', columns=['threshold_at_precision_0.99'])
                print(thr_at_p_99_pd)

                # ap info
                print()
                print('---------------')
                print(df_notnull)
                mAP = df_notnull.mean()['AP']
                print('------------')
                print(' mAP: {:.3f}'.format(mAP))
                print('wmAP: {:.3f}\n'.format(wmAP))

        ap_result = {
            'AP': ap_info,
            'wmAP': wmAP
        }

        self.ap_result = ap_result
        return ap_result

    def _set_threshold(self, policy='inflection', value=None):
        """
        Args:
            policy (str, optional): could be one of 'inflection' or 'manual'. Defaults to 'inflection'.
            value (optional): 
                under 'inflection' policy: value is insignificant
                under 'recall' or 'precision': value should be the minimum float we can accept
                under 'manual': value can be a list or dict who has the same length or keys with classnames. 
                . Defaults to None.
        """
        assert policy in ['inflection', 'manual'], f"policy {policy} should in ['inflection', 'recall', 'precision', 'manual']"
        threshold_dict = dict()

        if policy != 'manual':
            if not hasattr(self, 'ap_result'):
                print("@@ Warning: You did not run eval_ap manually, and we have to run it with default args.")
                self.eval_ap(verbose=False)
            assert hasattr(self, 'ap_result'), f"inflection threshold need self.ap_result, and we can not find it."

        if policy == 'inflection':
            ap_res = self.ap_result['AP']
            for key, ap_data in ap_res.items():
                inflection = ap_data['inflection']['inflection_score']

                threshold_dict[key] = inflection

        elif policy == 'manual':
            if isinstance(value, dict):
                threshold_dict = value
            else:
                assert isinstance(value, list), f"under manual threshold policy, value should be dict or list. {value}"
                assert len(value) == len(self.class_list), f"value list is not as long as class_list {len(value)} != {len(self.class_list)}"
                for index in range(len(self.class_list)):
                    threshold_dict[self.class_list[index]] = value[index]
        else:
            raise RuntimeError(f"unknown policy {policy}")

        self.threshold_dict = threshold_dict
        return threshold_dict

    def eval_judgment(self):
        if not hasattr(self, 'threshold_dict'):
            print("@@ Warning: You did not run set_threshold manually, and we have to run it with default args.")
            self.set_threshold()
        assert hasattr(self, 'threshold_dict'), f"eval_judgment failed for threshold_dict not found."
        pred_score_array = np.array(self.pred_data_list)
        thre_array = np.array([self.threshold_dict[key] for key in self.class_list])
        
        judge_res_list = (pred_score_array > thre_array).astype('uint8').tolist()
        self.judge_res_list = judge_res_list

    def get_confusion_matrix(self):
        if not hasattr(self, 'judge_res_list'):
            print("@@ Warning: You did not run eval_judgment manually, and we have to run it with default args.")
            self.eval_judgment()
        assert hasattr(self, 'judge_res_list'), f"get_confusion_matrix failed for judge_res_list not found."
        cm_dict = {key:dict(TP=0, TN=0, FN=0, FP=0) for key in self.class_list}

        y_true = np.array(self.gt_data_list).astype('bool').astype('uint8')
        y_pred = np.array(self.judge_res_list).astype('bool').astype('uint8')

        TP = (y_true * y_pred).sum(axis=0)
        TN = ((1 - y_true) * (1 - y_pred)).sum(axis=0)
        FP = ((1 - y_true) * y_pred).sum(axis=0)
        FN = (y_true * (1 - y_pred)).sum(axis=0)

        for index, key in enumerate(self.class_list):
            cm_dict[key]['TP'] = TP[index]
            cm_dict[key]['TN'] = TN[index]
            cm_dict[key]['FP'] = FP[index]
            cm_dict[key]['FN'] = FN[index]

        for key, cm_info in cm_dict.items():
            assert cm_info['TP'] + cm_info['TN'] + cm_info['FP'] + cm_info['FN'] == len(self.judge_res_list), f"cm_info {cm_info} sum error."
        
        # cm_list = multilabel_confusion_matrix(y_true=y_true, y_pred=y_pred)
        cm_pd = pd.DataFrame(cm_dict)
        return cm_pd

    def dump_failure_case(self, data_root, target_dir):
        if not hasattr(self, 'judge_res_list'):
            print("@@ Warning: You did not run eval_judgment manually, and we have to run it with default args.")
            self.eval_judgment()
        assert hasattr(self, 'judge_res_list'), f"dump_failure_case failed for judge_res_list not found."

        dir_check(target_dir)
        target_dir = Path(target_dir)

        for judge_res, gt_data, data_path in tqdm(zip(self.judge_res_list, self.gt_data_list, self.data_path_list), total=len(self.gt_data_list)):
            if not isinstance(data_path, str) and np.iterable(data_path):
                need_puzzle = True
                assert len(data_path) > 0, f"empty path list."
            else:
                need_puzzle = False

            for index, label in enumerate(self.class_list):
                pred_label = judge_res[index]
                gt_label = gt_data[index]
                
                if pred_label == gt_label:
                    continue
                else:
                    if gt_label == 0:
                        mode = 'false_alarm'
                    elif gt_label == 1:
                        mode = 'miss'
                    else:
                        raise ValueError(f"bad gt label value {gt_label}")

                    # save image name
                    if need_puzzle:
                        ori_img_path = Path(data_root) / data_path[0]
                    else:
                        ori_img_path = Path(data_root) / data_path
                    
                    # image load
                    if need_puzzle:
                        # image_list = [ cv_rgb_imread(Path(data_root) / path) for path in data_path]
                        ######################
                        image_list = list()
                        for image_path in data_path:
                            data_path = str(Path(data_root) / image_path).replace('\\', '/')
                            try:
                                image = cv_rgb_imread(data_path)
                            except:
                                try:
                                    image = cv_rgb_imread(data_path.replace('attention/normal', 'attention/dark_field'))
                                except:
                                    image = cv_rgb_imread(data_path.replace('attention/normal', 'attention/large_image'))
                            image_list.append(image)
                        #####################
                        image = puzzle(image_list)
                    else:
                        if not ori_img_path.exists():
                            print(f"Ori image path {ori_img_path} not found.")
                            continue
                        else:
                            try:
                                image = cv_rgb_imread(ori_img_path)
                            except Exception as e:
                                print(f"image load failed! {ori_img_path}")
                                print(f"error message: {e}")
                                continue

                    save_path = target_dir / mode / label / encode_path(ori_img_path)
                    cv_rgb_imwrite(image, save_path)

    eval = eval_ap
    evaluate = eval_ap
    evaluation = eval_ap


ClassificationEvaluator = ClassificationMultiLabelEvaluator
MultiLabelEvaluator = ClassificationMultiLabelEvaluator


# -*- coding: utf-8 -*-
# @Author: Zhai Menghua
# @Date:   2020-07-16 10:10:48
# @Last Modified by:   Zhai Menghua
# @Last Modified time: 2020-07-16 10:13:00

# -*-coding:utf-8-*-
#
# Compress and Decompress the shape object
#
import numpy as np


def encode_distribution(distribution):
    """
    Encode the distribution (usually a list of float)
    Return a serializable object (usually a string) for json dumping
    """
    code = list()
    for data in distribution:
        if None: pass
        elif isinstance(data, (float, int, np.float16, np.float32, np.int32, np.int64)):
            if data in [0,1]:
                code.append(str(data))
            else:
                code.append('{:0.6e}'.format(data))
        elif isinstance(data, int):
            code.append(str(data))
        else:
            raise RuntimeError
    code = ','.join(code)
    return code


def decode_distribution(encoded_distribution):
    """
    Decode the cnn json distribution (usually encoded data)
    Return a list of float
    """
    if isinstance(encoded_distribution, list):
        return encoded_distribution
    assert isinstance(encoded_distribution, str)
    distribution_str_list = encoded_distribution.split(',')
    distribution = list()
    for data_str in distribution_str_list:
        data = float(data_str)
        distribution.append(data)
    return distribution


def encode_labelme_shape(point_list):
    """
    Encode the labelme shape (usually a list of points)
    Return a serializable object (usually a string) for json dumping
    """
    code = list()
    for point in point_list:
        assert len(point) == 2
        code.append('{:.6f}+{:.6f}'.format(point[0], point[1]))
    code = ','.join(code)
    return code


def decode_labelme_shape(encoded_shape):
    """
    Decode the cnn json shape (usually encoded from labelme format)
    Return a list of points that are used in labelme
    """
    assert isinstance(encoded_shape, str)
    points = encoded_shape.split(',')
    shape = list()
    for point in points:
        x, y = point.split('+')
        shape.append([float(x), float(y)])
    return shape



# 该代码由sklearn官方提供

import numpy as np
import matplotlib.pyplot as plt
from sklearn import svm, datasets
from sklearn.model_selection import train_test_split
from sklearn.metrics import confusion_matrix
from sklearn.utils.multiclass import unique_labels


def plot_confusion_matrix(y_true, y_pred, classes, normalize=False, title=None, cmap=plt.cm.Blues):
    """该函数用于画混淆矩阵。
    Args:
        y_true: 真实标签，1维array，如np.array([0, 3, 3, 2, 4, 2, 1, 3])
        y_pred: 预测标签，1维array, 如np.array([0, 2, 3, 2, 4, 2, 0, 3])
        classes: 标签名数组，1维np.array, 不能为list, e.g., np.array(['ok', 'ng1', 'ng2', 'ng3', 'ng4'])
        normalize: True：输出的为具体分对和分错的样本个数，False：输出的为分对和分错的比例
        title：混淆矩阵标题
        cmap：颜色
    """
    if not title:
        if normalize:
            title = 'Normalized confusion matrix'
        else:
            title = 'Confusion matrix, without normalization'

    # Compute confusion matrix
    cm = confusion_matrix(y_true, y_pred)
    # Only use the labels that appear in the data
    classes = classes[unique_labels(y_true, y_pred)]
    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
        print("Normalized confusion matrix")
    else:
        print('Confusion matrix, without normalization')

    fig, ax = plt.subplots()
    im = ax.imshow(cm, interpolation='nearest', cmap=cmap)
    ax.figure.colorbar(im, ax=ax)
    # We want to show all ticks...
    ax.set(xticks=np.arange(cm.shape[1]),
           yticks=np.arange(cm.shape[0]),
           # ... and label them with the respective list entries
           xticklabels=classes, yticklabels=classes,
           title=title,
           ylabel='True label',
           xlabel='Predicted label')

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
             rotation_mode="anchor")

    # Loop over data dimensions and create text annotations.
    fmt = '.2f' if normalize else 'd'
    thresh = cm.max() / 2.
    for i in range(cm.shape[0]):
        for j in range(cm.shape[1]):
            ax.text(j, i, format(cm[i, j], fmt),
                    ha="center", va="center",
                    color="white" if cm[i, j] > thresh else "black")
    fig.tight_layout()
    return ax


if __name__ == '__main__':
    
    # y_true = [0, 3, 3, 2, 4, 2, 1, 3]
    # y_pred = [0, 2, 3, 2, 4, 2, 0, 3]
    # class_names = np.array(['cls1', 'cls2', 'cls3', 'cls4', 'cls5'])
    # plot_confusion_matrix(np.array(y_true), np.array(y_pred), class_names,
    #                       normalize=False, title='hehe')
    # plt.show()

    cls1 = np.ones(15447) * 0
    cls2 = np.ones(813) * 1
    cls3 = np.ones(266) * 2
    cls4 = np.ones(909) * 3
    cls5 = np.ones(379) * 4
    cls6 = np.ones(60) * 5

    y_true = np.concatenate([cls1, cls2, cls3, cls4, cls5, cls6])
    print(y_true.shape)


import numpy as np

def label_to_onehot(label_list, nb_classes):
    """将一个2维的label index list转为2维的onehot array 
    Args:
        label_list: 一个2维的label list, e.g., [[2], [0, 1], [1]]
        nb_classes: 类别数量, e.g., 3
    Return:
        一个二维的array, e.g., np.array([[0, 0, 1], [1, 1, 0], [0, 1, 0]])
    """
    res_arr = np.zeros([len(label_list), nb_classes])
    for i, label in enumerate(label_list):
        res_arr[i][label] = 1
    return res_arr

def onehot_to_label(onehot_array):
    """将一个2维的onehot array 转为1维的label list， 支持多标签形式
    Args:
        一个二维的array, e.g., [[0, 0, 1], [1, 0, 0], [0, 1, 1]]
    Return:
        label_list: 一个2维的label list, e.g., [[2], [0], [1, 2]]
    """
    label_list = []
    for arr in onehot_array:
        label_list.append(list(np.where(arr == 1)[0]))
    return label_list


if __name__ == '__main__':
    a = np.array([[0, 0, 1, 0], [1, 0, 1, 0], [0, 0, 0, 1]])
    b = onehot_to_label(a)
    print(b)
    c = label_to_onehot(b, 4)
    print(c)


import copy
import numpy as np

from .base import DataManagerBase
from ..utils import encode_distribution
from ..utils import decode_distribution


class DataManagerFormater(DataManagerBase):
    def __init__(self, record_list=[], class_list=[]):
        assert isinstance(record_list, list)
        assert isinstance(class_list,  list)
        self.class_list  = self._extract_leaf_classes(class_list)
        self.record_list = self._format_record_list(record_list)


    def _format_distribution(self, old_distribution):
        """
        Format distribution/scores from legacy detection/multiclass/multilabel data
        The legacy data contains parent classes, which are incompatible with current
        version of dataset, we need to fix that 

        Args:
            distribution: 1D numpy array or list

        Return:
            a new distribution aligned with self.class_list
        """
        _old_class2label = {x['class_name']:x['class_id'] for x in self._old_class_dict}
        new_distribution = list()
        for classname in self.class_list:
            classid = _old_class2label[classname]
            new_distribution.append(old_distribution[classid])
        return new_distribution


    def _format_record(self, record, target_task=None):
        """
        Format record

        Args:
            record: you want to format
            target_task: task type you want format the record to (None: keep the same)

        Return:
            A reformated record
        """
        source_type = self._get_record_task_type(record)
        target_task = source_type.split('-')[0] if target_task is None else target_task
        if source_type == 'detection-legacy':
            # 1) covert standard detection task
            new_record = copy.deepcopy(record)
            for inst in new_record['instances']:
                old_distribution = decode_distribution(inst.pop('distribution'))
                distribution = self._format_distribution(old_distribution)
                classid = np.argmax(distribution)
                score, classname = distribution[classid], self.class_list[classid]
                inst.update({
                    'score': score,
                    'classname': classname,
                })

            # 2) convert to target task type
            if target_task == 'detection':
                return new_record
            elif target_task == 'multilabel':
                instances  = new_record.pop('instances')
                new_scores = [0] * len(self.class_list)
                for inst in instances:
                    classname, score = inst['classname'], inst['score']
                    classid = self.class2label(classname)
                    new_scores[classid] = max(score, new_scores[classid])
                new_record['scores'] = encode_distribution(new_scores)
            else:
                raise RuntimeError("Can't format record from task '{}' to '{}'".format(source_type, target_task))

        elif source_type == 'detection':
            new_record = copy.deepcopy(record)
            if target_task == 'detection':
                for inst in new_record['instances']:
                    if 'class_name' in inst:
                        inst['classname'] = inst.pop('class_name')
            elif target_task == 'multilabel':
                instances = new_record.pop('instances')
                new_scores = [0] * len(self.class_list)
                for inst in instances:
                    classname, score = inst['classname'], inst['score']
                    classid = self.class2label(classname)
                    new_scores[classid] = max(score, new_scores[classid])
                new_record['scores'] = encode_distribution(new_scores)
            else:
                raise RuntimeError("Can't format record from task '{}' to '{}'".format(source_type, target_task))

        elif source_type == 'multiclass-legacy':
            if target_task == 'multiclass':
                new_record = copy.deepcopy(record)
                old_distribution = decode_distribution(new_record['info'].pop('distribution'))
                distribution = self._format_distribution(old_distribution)
                new_record['distribution'] = encode_distribution(distribution)
            else:
                raise RuntimeError("Can't format record from task '{}' to '{}'".format(source_type, target_task))

        elif source_type == 'multiclass':
            if target_task == 'multiclass':
                new_record = copy.deepcopy(record)
            else:
                raise RuntimeError("Can't format record from task '{}' to '{}'".format(source_type, target_task))

        elif source_type == 'multilabel-ancient':
            if target_task == 'multilabel':
                new_record = copy.deepcopy(record)
                instances = new_record.pop('instances')
                assert len(instances) in [0, 1], instances
                new_record['scores'] = [0] * len(self.class_list)
                for inst in instances:
                    old_scores = decode_distribution(inst.pop('distribution'))
                    new_scores = self._format_distribution(old_scores)
                    new_record['scores'] = encode_distribution(new_scores)
            else:
                raise RuntimeError("Can't format record from task '{}' to '{}'".format(source_type, target_task))

        elif source_type == 'multilabel-legacy':
            if target_task == 'multilabel':
                new_record = copy.deepcopy(record)
                old_distribution = decode_distribution(new_record['info'].pop('scores'))
                distribution = self._format_distribution(old_distribution)
                new_record['scores'] = encode_distribution(distribution)
            else:
                raise RuntimeError("Can't format record from task '{}' to '{}'".format(source_type, target_task))

        elif source_type == 'multilabel':
            if target_task == 'multilabel':
                new_record = copy.deepcopy(record)
            else:
                raise RuntimeError("Can't format record from task '{}' to '{}'".format(source_type, target_task))

        else:
            raise RuntimeError("Unrecognized record task: {}".format(source_type))

        return new_record

    def _format_record_list(self, record_list, target_task=None):
        """
        Format record list

        Args:
            record_list: list of record
            target_task: task type you want format the record to

        Return:
            A reformated record list
        """
        new_record_types = [self._format_record(rec, target_task) for rec in record_list]
        return new_record_types

    def _extract_leaf_classes(self, class_list):
        self._old_class_dict = class_list

        leaf_class_list = list()
        for x in class_list:
            if isinstance(x, dict): # this indicates that we are dealing with the lagecy dataset
                nonleaf_classes = set([x['parent'] for x in class_list if x['parent'] is not None])
                if x['class_name'] not in nonleaf_classes:
                    leaf_class_list.append(x['class_name'])
            elif isinstance(x, str):
                return class_list
            else:
                raise TypeError("Unrecognized class list")
        return leaf_class_list

    def _get_record_task_type(self, record):
        """
        Return the type of the given record
        """
        assert isinstance(record, dict)
        assert 'info' in record

        # detection record
        if 'instances' in record:
            modes = set()
            instances = record['instances']
            if len(instances) > 0:
                for inst in instances:
                    if 'distribution' in inst:
                        assert 'shape_type' in inst, inst
                        if inst['shape_type'] is None:
                            modes.add('multilabel-ancient')
                        else:
                            modes.add('detection-legacy')
                    else:
                        modes.add('detection')
                        assert 'score' in inst, "Unknown instances: {}".format(inst)

            if len(modes) == 0:
                return 'detection'
            else:
                assert len(modes) == 1, "Unknown detection type: {}".format(record)
                return modes.pop()

        # classification record
        if 'scores' in record['info']:
            return 'multilabel-legacy'
        elif 'scores' in record:
            return 'multilabel'
        elif 'distribution' in record['info']:
            return 'multiclass-legacy'
        elif 'distribution' in record:
            return 'multiclass'
        else:
            raise TypeError('Unknown record type: {}'.format(record))

    def _det2ml(self):
        data_task_type = self.task_type
        assert data_task_type == 'detection', \
            "Only detection dataset can be converted to multilabel dataset: {}".format(data_task_type)
        new_record_list = []
        for rec in self.record_list:
            scores = [0]*len(self.class_list)
            for inst in rec['instances']:
                classid = self.class2label(inst['classname'])
                scores[classid] = max(scores[classid], inst['score'])
            new_record = {
                'info': copy.deepcopy(rec['info']),
                'scores': encode_distribution(scores),
            }
            new_record_list.append(new_record)
        return type(self)(record_list=new_record_list, class_list=self.class_list)

    def _ml2mc(self):
        data_task_type = self.task_type
        assert data_task_type == 'multilabel'
        print(
            "Warning: Convert from Multilabel to Multiclass is not legitimate, \n"
            "We adopt a simple strategy for doing that: \n"
            "multiclass distribution = multilabel scores / sum(multilabel scores)"
        )
        new_record_list = []
        for rec in self.record_list:
            scores = self.get_multilabel_scores_from_record(rec)
            distribution = scores / (np.sum(scores) + 1e-12)
            new_record = {
                'info': copy.deepcopy(rec['info']),
                'distribution': encode_distribution(distribution),
            }
            new_record_list.append(new_record)
        return type(self)(record_list=new_record_list, class_list=self.class_list)

    def _mc2ml(self):
        data_task_type = self.task_type
        assert data_task_type == 'multiclass'
        print(
            "Warning: Convert from Multiclass to Multilabel is not legitimate, \n"
            "We adopt a simple strategy for doing that: \n"
            "multilabel scores = multiclass distribution"
        )
        new_record_list = []
        for rec in self.record_list:
            distribution = self.get_multiclass_distribution_from_record(rec)
            new_record = {
                'info': copy.deepcopy(rec['info']),
                'scores': encode_distribution(distribution),
            }
            new_record_list.append(new_record)
        return type(self)(record_list=new_record_list, class_list=self.class_list)

    def to_ml(self):
        data_task_type = self.task_type
        if data_task_type == 'detection':
            return self._det2ml()
        elif data_task_type == 'multilabel':
            return self.clone()
        else:
            return self._mc2ml()

    def to_mc(self):
        data_task_type = self.task_type
        if data_task_type == 'detection':
            data_ml = self._det2ml()
            return data_ml.to_mc()
        elif data_task_type == 'multilabel':
            return self._ml2mc()
        else:
            return self.clone()



#
# Class for labelme json management
#
import json
from json import encoder
import os, cv2, copy
import os.path as osp
import json
from ..utils import MyEncoder

from .polygon import Polygon

class Labelme(object):
    def __init__(self, image_info, shapes):
        self.shape_list = shapes
        self.image_info = image_info
        self.image_path = image_info['image_path']
        self.image_height = image_info['height']
        self.image_width = image_info['width']

    @classmethod
    def from_json(cls, labelme_json):
        labelme_json = str(labelme_json)
        data = cls.parse_json(labelme_json)
        return cls(image_info=data['image_info'], shapes=data['shapes'])

    def __len__(self):
        return len(self.shape_list)

    def __getitem__(self, idx):
        return self.shape_list[idx]

    def iterator(self):
        for shape in self.shape_list:
            yield shape

    def clone(self):
        return copy.deepcopy(self)

    @staticmethod
    def parse_json(labelme_json):
        try:
            with open(labelme_json, encoding='gbk') as fid:
                obj = json.load(fid)
        except:
            with open(labelme_json, encoding='utf-8') as fid:
                obj = json.load(fid)

        # get shapes
        shapes = list()
        for shape in obj['shapes']:
            polygon = Polygon(
                points=shape['points'],
                class_name=shape['label'],
                shape_type=shape['shape_type']
            )
            shapes.append(polygon)

        image_info = {
            'image_path': obj['imagePath'].replace('\\', '/'),  # path to linux style
            'height': obj['imageHeight'], 'width': obj['imageWidth']
        }
        data = {'shapes': shapes, 'image_info':  image_info}

        return data

    def load_image(self, image_root=''):
        image_path = os.path.join(image_root, self.image_path)
        assert os.path.exists(image_path), "Cannot find image: {}".format(image_path)
        image_bgr = cv2.imread(image_path)
        image_rgb = copy.deepcopy(image_bgr[...,::-1]) # deepcopy to get around the polyline drawing error
        return image_rgb

    def draw_shapes(self, image, linewidth=20, color=(255,0,0)):
        """ draw shapes on image """
        for shape in self.iterator():
            image = shape.draw_on(image, linewidth=linewidth, color=color)
        return image

    def save_json(self, json_path):
        json_dir = osp.dirname(json_path)
        os.makedirs(json_dir, exist_ok=True)

        obj = {
            'version': '4.4.0',
            'flags': {},
            'shapes': [shp.json_format() for shp in self.shape_list],
            'imagePath': self.image_path,
            'imageData': None,
            'imageHeight': self.image_height,
            'imageWidth': self.image_width
        }
        
        with open(json_path, 'w') as fid:
            json.dump(obj, fid, indent=4, cls=MyEncoder)



import copy
import numpy as np

from .info import DataManagerInfoExtractor
from ..utils import create_uuid
from ..utils import encode_distribution
from ..utils import encode_labelme_shape


class DataManagerSaver(DataManagerInfoExtractor):
    def _format_shape_to_polygon(self, shape, shape_type):
        assert isinstance(shape, (list, tuple, np.ndarray))
        assert len(shape) >= 2
        if None: pass
        elif shape_type == 'rectangle':
            assert len(shape) == 2, "shape as rectangle example: [(x_topleft,y_topleft), (x_bottomrignt, y_bottomright)]"
            assert isinstance(shape[0], (list, tuple, np.ndarray))
            (x1,y1),(x2,y2) = shape[0], shape[1]
            points = [[x1,y1],[x1,y2],[x2,y2],[x2,y1],[x1,y2]]
        elif shape_type == 'xyxy':
            assert len(shape) == 4, "shape as xyxy example: [x1, y1, x2, y2]"
            x1,y1,x2,y2 = shape
            points = [[x1,y1],[x1,y2],[x2,y2],[x2,y1],[x1,y2]]
        elif shape_type == 'xywh':
            assert len(shape) == 4, "shape as xyxy example: [x, y, w, h]"
            x,y,w,h = shape
            x1,y1,x2,y2 = x,y,x+w,y+h
            points = [[x1,y1],[x1,y2],[x2,y2],[x2,y1],[x1,y2]]
        elif len(shape) >= 3:
            points = [[x, y] for x, y in shape]
        else:
            raise RuntimeError("Invalid shape: {}".format(shape))
        return points

    def create_multilabel_scores_from_classid(self, classid, score=1.):
        classname = self.label2class(classid)
        return self.create_multilabel_scores_from_classname(classname)

    def create_multilabel_scores_from_classname(self, classname, score=1.):
        scores = [0] * len(self.class_list)
        classid = self.get_classid_from_classname(classname)
        scores[classid] = score
        return encode_distribution(scores)

    def create_multilabel_scores_from_prediction(self, prediction):
        """
        Create encoded scores from class prediction
        """
        scores = [0] * len(self.class_list)
        assert len(prediction) == len(scores)
        for classid, pred_score in enumerate(prediction):
            scores[classid] = pred_score
        return encode_distribution(scores)

    def create_multiclass_distribution_from_classid(self, classid, score=1.):
        return self.create_multilabel_scores_from_classid(classid, score)

    def create_multiclass_distribution_from_classname(self, classname, score=1.):
        return self.create_multilabel_scores_from_classname(classname, score)

    def create_multiclass_distribution_from_prediction(self, prediction):
        assert all(0 <= sc <= 1 for sc in prediction)
        assert abs(1 - np.sum(prediction)) < 1e-5, "prob distribution should be sum to 1: {}".format(np.sum(prediction))
        return self.create_multilabel_scores_from_prediction(prediction)

    def create_detection_instance_from_classname(self, classname, score=1., shape=[-1,-1,-1,-1], shape_type='xyxy'):
        # create class id
        assert classname in self.class_list
        points = self._format_shape_to_polygon(shape, shape_type)
        instance = {
            'uuid': create_uuid(),
            'classname': classname,
            'score': score,
            'shape_type': 'polygon',
            'points': encode_labelme_shape(points)
        } 
        return instance

    def create_detection_instance_from_classid(self, classid, score=1., shape=[-1,-1,-1,-1], shape_type='xyxy'):
        classname = self.label2class(classid)
        return self.create_detection_instance_from_classname(classname, score, shape, shape_type)



"""该文件包含二分类、多类别分类和多标签分类的评估代码。"""

from .utils import onehot_to_label, label_to_onehot
from sklearn import metrics
import matplotlib.pyplot as plt
import matplotlib.image as mpimg
import numpy as np
import os
import cv2
import warnings
import sys


class ClassifierEval(object):
    """A class to evaluate the performance of a classifier."""

    @classmethod
    def compute_ap(cls):
        """Compute AP of each class."""
        pass

    @classmethod
    def compute_p_at_r(cls):
        """Compute precision at (Recall >= referred_recall)."""
        pass

    @classmethod
    def draw_pr_curve(cls):
        """Draw precision-recall curve."""
        pass

    @classmethod
    def compute_fnr_and_fpr(cls):
        """Refer to https://en.wikipedia.org/wiki/Receiver_operating_characteristic for detailed calculation
        fnr = 1 - tpr = fn / (tp + fn), miss rate (漏报率)
        fpr = fp / (tn + fp), false alarm rate (误报率)
        """
        pass


class ClassifierEvalBinary(ClassifierEval):
    """二分类评估器。"""

    @classmethod
    def compute_ap(cls, y_true, y_score):
        """计算二分类的AP 
        Args:
            y_true: 1维nparray, 所有样本的真值列表[nSamples], e.g., [0, 1, 1, 0], 1为ng, 0为ok
            y_score: 1维nparray, 所有样本在目标类别上的分数, 注意该分数是最后一层(softmax层)的输出结果。e.g., [0.2, 0.9, 0.4, 0.5]
        Return:
            一个数字:AP值。
        """
        if np.sum(y_true) <= 0:
            ap = 0
        else:
            ap = metrics.average_precision_score(y_true=y_true, y_score=y_score)
        return round(ap, 6)

    @classmethod
    def compute_p_at_r(cls, y_true, y_score, recall_thresh=0.995):
        """计算二分类中recall >= 0.995时的precision。
        Args:
            y_true: 1维nparray, 所有样本的真值列表[nSamples], 以0和1形式给出, e.g., [0, 0, 1, 0]
            y_score: 1维nparray, 所有样本在正类上的分数, e.g., [0.2, 0.9, 0.7, 0.1]
        Return:
            Precision@(Recall>=recall_thresh)
        """
        if np.any(y_true):
            precision, recall, _ = metrics.precision_recall_curve(y_true, y_score)
            ind_all = np.where(recall < recall_thresh)
            ind = ind_all[0][0] - 1
            return round(precision[ind], 4)
        else:
            return 0.

    @classmethod
    def draw_pr_curve(cls, y_true, y_score, output_path='./pr_curve.png', class_name=""):
        """画二分类的pr曲线。
        Args:
            y_true: 1维nparray, 所有样本的真值列表[nSamples], 以0和1形式给出, e.g., [0, 0, 1, 0]
            y_score: 1维nparray, 所有样本在正类上的分数, e.g., [0.2, 0.9, 0.7, 0.1]
        Return:
            无
        """
        dirname = os.path.dirname(output_path)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        #precision, recall, _ = metrics.precision_recall_curve(y_true, y_score)

        labels = list(y_true)
        scores = list(y_score)
        total_yes = sum(labels)

        # sort scores in descend order
        bundle = sorted(zip(scores, labels), key=lambda x: -x[0])

        recall_list, precision_list = list(), list()
        yes_counter = 0
        if total_yes > 0:
            for ix, (_, lb) in enumerate(bundle):
                all_counter = ix+1
                yes_counter += lb  # when lb = 1, it yelps "YES!"
                recall = yes_counter / total_yes
                precision = yes_counter/all_counter
                recall_list.append(recall)
                precision_list.append(precision)

        # draw pr curve
        plt.figure()
        #plt.plot(recall, precision, '.')
        plt.step(recall_list, precision_list, where='pre')
        plt.xlabel('Recall')
        plt.ylabel('Precision')
        plt.ylim([0.0, 1.05])
        plt.xlim([0.0, 1.0])

        ap = cls.compute_ap(y_true, y_score)

        plt.title(str(class_name)+' PR Curve - AP is {:.6f}'.format(ap))

        plt.savefig(output_path)
        plt.close()

        return ap

    @classmethod
    def get_inflection_info(cls, y_true, y_score):
        """
        计算拐点信息与ap信息
        args:
            y_true: 1维nparray, 所有样本的真值列表[nSamples], 以0和1形式给出, e.g., [0, 0, 1, 0]
            y_score: 1维nparray, 所有样本在正类上的分数, e.g., [0.2, 0.9, 0.7, 0.1]
        returns:
            ap_info_dict: 包含拐点与AP信息的字典
        """
        labels = list(y_true)
        scores = list(y_score)
        total_yes = sum(labels)

        # sort scores in descend order
        bundle = sorted(zip(scores, labels), key=lambda x: -x[0])

        recall_list, precision_list = list(), list()
        yes_counter = 0

        def distance_square_cal(recall, precision):
            return recall * precision

        inflection_recall = 0
        inflection_precision = 0
        inflection_score = 1
        max_distance_square = 0
        if total_yes > 0:
            for ix, (score, lb) in enumerate(bundle):
                all_counter = ix+1
                yes_counter += lb  # when lb = 1, it yelps "YES!"
                recall = yes_counter / total_yes
                precision = yes_counter/all_counter
                distance = distance_square_cal(recall, precision)
                if distance >= max_distance_square:
                    max_distance_square = distance
                    inflection_recall = recall
                    inflection_precision = precision
                    inflection_score = score
                recall_list.append(recall)
                precision_list.append(precision)

        ap = cls.compute_ap(y_true, y_score)

        ap_info_dict = dict()
        ap_info_dict['inflection_score'] = inflection_score
        ap_info_dict['inflection_recall'] = inflection_recall
        ap_info_dict['inflection_precision'] = inflection_precision
        ap_info_dict['max_distance_square'] = max_distance_square
        ap_info_dict['ap'] = ap

        return ap_info_dict

    @classmethod
    def compute_fnr_and_fpr(cls, y_true, y_score, fnr_thresh=0.005, fail_study=False):
        """计算二分类中的漏报率和误报率, 要求漏报率必须小于等于fnr_thresh。
        Args:
            y_true: 1维nparray, 所有样本的真值列表[nSamples], 以0和1形式给出, e.g., [0, 0, 1, 0]
            y_score: 1维nparray, 所有样本在正类上的分数, e.g., [0.2, 0.9, 0.7, 0.1]
            fnr_thresh: 能容忍的最大的漏报率
            fail_study: True时返回失败案例的索引, 含漏报和误报的。False时返回空list。
        Return:
            min_score: 判断为ng的最小score, ng类别的阈值大于等于该score则判断为ng, 否则判断为ok 
            res: 含fnr和fpr
            fn_index_list: fail_study为False时为[], 否则为漏报的图片的index list
            fp_index_list: fail_study为False时为[], 否则为误报的图片的index list
        """
        fn_index_list, fp_index_list = [], []
        _, recall, thresholds = metrics.precision_recall_curve(y_true, y_score)
        ind_all = np.where(recall < (1 - fnr_thresh))
        ind = ind_all[0][0] - 1
        min_score = thresholds[ind]
        y_pred = [int(per_score >= min_score) for per_score in y_score]
        tp, fp, tn, fn = 0, 0, 0, 0
        if fail_study:
            for i, (per_true, per_pred) in enumerate(zip(y_true, y_pred)):
                if per_true == 1 and per_pred == 1:
                    tp += 1
                elif per_true == 1 and per_pred == 0:
                    fn += 1
                    fn_index_list.append(i)
                elif per_true == 0 and per_pred == 1:
                    fp += 1
                    fp_index_list.append(i)
                elif per_true == 0 and per_pred == 0:
                    tn += 1
        else:
            for per_true, per_pred in zip(y_true, y_pred):
                if per_true == 1 and per_pred == 1:
                    tp += 1
                elif per_true == 1 and per_pred == 0:
                    fn += 1
                elif per_true == 0 and per_pred == 1:
                    fp += 1
                elif per_true == 0 and per_pred == 0:
                    tn += 1
        fnr = fn / (tp + fn)
        fpr = fp / (tn + fp)
        res = dict()
        res['fnr'] = round(fnr, 4)
        res['fpr'] = round(fpr, 4)
        return min_score, res, fn_index_list, fp_index_list

    @classmethod
    def draw_failure_cases(cls, img_path_list, y_true, y_score, min_score, fn_index_list, fp_index_list, cls_dict, res_dir=None):
        """保存failure cases, 含漏报和误报的。
        Args:
            img_path_list: 图片路径
            y_true: 1维nparray, 所有样本的真值列表[nSamples], 以0和1形式给出, e.g., [0, 0, 1, 0]
            y_score: 1维nparray, 所有样本在正类上的分数, e.g., [0.2, 0.9, 0.7, 0.1]
            min_score: 判断为ng的最小score值
            fn_index_list: 漏报的样本索引
            fp_index_list: 误报的样本索引
            cls_dict: 类别字典, 用于做failure case的显示用, 如{0:'OK', 1:'NG'}
            res_dir: failure case存放路径
        Return:
            无
        """
        # Step 1. Make dirs.
        if not res_dir:
            raise ValueError('Result directory error!')
        res_dir_fn = os.path.join(res_dir, 'loubao')
        res_dir_fp = os.path.join(res_dir, 'wubao')
        if not os.path.exists(res_dir_fn):
            os.makedirs(res_dir_fn)
        if not os.path.exists(res_dir_fp):
            os.makedirs(res_dir_fp)

        # Step 2. Get dt_label_list.
        dt_label_list = []
        for per_score in y_score:
            if per_score >= min_score:
                dt_label_list.append(1)
            else:
                dt_label_list.append(0)

        # Step 3. Draw failures.
        if not fn_index_list:
            print('fn_index_list is empty!')
        else:
            for ind in fn_index_list:
                gt_label = y_true[ind]
                dt_label = dt_label_list[ind]
                img = mpimg.imread(img_path_list[ind])
                plt.imshow(img)
                plt.title(
                    'cls: {} -> cls: {}'.format(cls_dict[gt_label], cls_dict[dt_label]))
                plt.axis('off')
                output_path = os.path.join(
                    res_dir_fn, os.path.basename(img_path_list[ind]))
                output_path_new = output_path.replace('.jpg', '.png')
                plt.savefig(output_path_new)
                plt.close()

        if not fp_index_list:
            print('fp_index_list is empty!')
        else:
            for ind in fp_index_list:
                gt_label = y_true[ind]
                dt_label = dt_label_list[ind]
                img = mpimg.imread(img_path_list[ind])
                plt.imshow(img)
                plt.title(
                    'cls: {} -> cls: {}'.format(cls_dict[gt_label], cls_dict[dt_label]))
                plt.axis('off')
                output_path = os.path.join(
                    res_dir_fp, os.path.basename(img_path_list[ind]))
                output_path_new = output_path.replace('.jpg', '.png')
                plt.savefig(output_path_new)
                plt.close()

    @classmethod
    def get_threshold_at_p(cls, y_true, y_score, precision_thresh=0.9):
        labels = list(y_true)
        scores = list(y_score)
        # sort scores in descend order
        bundle = sorted(zip(scores, labels), key=lambda x: -x[0])

        total_yes = sum(labels)
        recall_threshold = 0.05
        sample_num_thre = 6

        threshold = 1
        total_num = 0
        correct_num = 0
        for score, label in bundle:
            total_num += 1
            correct_num += label
            if correct_num/total_num > precision_thresh and correct_num/total_yes >= recall_threshold and correct_num >= sample_num_thre:
                threshold = score

        return threshold


class ClassifierEvalMulticlass(ClassifierEval):
    """多类别分类评估器。"""

    @classmethod
    def compute_ap(cls, y_true, y_score):
        """计算多类别分类的AP 
        Args:
            y_true: 2维nparray, 所有样本的真值列表[nSamples, nClasses], e.g., [2, 0]
            y_score: 2维nparray, 所有样本在各个类别上的分数, 注意每一个样本在各个类别上的分数之和须为1, e.g., [[0.2, 0.3, 0.4, 0.1], [0.1, 0.1, 0.3, 0.5], ...]
        Return:
            ap_list: 各个类别的AP值组成的list
        """
        # 1. 将类别转为one_hot的形式
        nb_classes = y_score.shape[1]
        y_true = label_to_onehot(y_true, nb_classes)
        # 2. 计算各个类别的ap
        ap_dict = dict()
        for i, (true_per_cls, score_per_cls) in enumerate(zip(y_true.T, y_score.T)):
            ap = metrics.average_precision_score(
                y_true=true_per_cls, y_score=score_per_cls)
            ap_dict[i] = round(ap, 4)
        return ap_dict

    @classmethod
    def compute_p_at_r(cls, y_true, y_score, recall_thresh=0.995):
        """计算多分类中recall >= 0.995时各个类别的precision。
        Args:
            y_true: 1维nparray, 所有样本的真值列表[nSamples], e.g., [2, 0, 3]
            y_score: 1维nparray, 所有样本在各个类上的分数, e.g., [[0.2, 0.7, 0.1, 0], [0.1, 0.3, 0.5, 0.1], [0.4, 0.2, 0.1, 0.3]]
        Return:
            每个类别的Precision@(Recall>=recall_thresh)组成的dict, 如{0: 0.2345, 1: 0.2344, 2: 0.7623, 3: 0.3334}
        """
        label_list = sorted(list(set(y_true)))  # label从小到大排序
        prec_dict = dict()
        for i, label in enumerate(label_list):
            y_true_per_cls = np.array([int(lab == label) for lab in y_true])
            y_score_per_cls = y_score[:, i]
            precision = ClassifierEvalBinary.compute_p_at_r(
                y_true_per_cls, y_score_per_cls, recall_thresh=recall_thresh)
            prec_dict[label] = round(precision, 4)  # 保留4位小数
        return prec_dict

    @classmethod
    def draw_pr_curve(cls, y_true, y_score, output_dir='hehe'):
        """画多类别分类的pr曲线, 每个类别一个pr曲线。
        Args:
            y_true: 1维nparray, 所有样本的真值列表[nSamples], e.g., [2, 0, 3]
            y_score: 1维nparray, 所有样本在各个类上的分数, e.g., [[0.2, 0.7, 0.1, 0], [0.1, 0.3, 0.5, 0.1], [0.4, 0.2, 0.1, 0.3]]
        Return:
            无
        """
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        label_list = sorted(list(set(y_true)))  # label从小到大排序
        for i, label in enumerate(label_list):
            y_true_per_cls = np.array([int(lab == label) for lab in y_true])
            y_score_per_cls = y_score[:, i]
            output_path = os.path.join(output_dir, 'cls_' + str(label))
            ClassifierEvalBinary.draw_pr_curve(
                y_true_per_cls, y_score_per_cls, output_path=output_path)

    @classmethod
    def compute_fnr_and_fpr(cls, y_true, y_score, ok_ind=0, fnr_thresh=0.005, fail_study=False):
        """计算多类别分类中的漏报率和误报率, 要求漏报率必须小于等于fnr_thresh。
        Args:
            y_true: 1维nparray, 所有样本的真值列表[nSamples], e.g., [2, 3, 1, 0]
            y_score: 2维nparray, 所有样本在各个类上的分数, e.g., [[0.2, 0.7, 0.1, 0], [0.1, 0.3, 0.5, 0.1], [0.4, 0.2, 0.1, 0.3], [0.3, 0.3, 0.4, 0]]
            ok_ind: ok类别的index, 默认0为ok, 其余为ng 
            fnr_thresh: 能容忍的最大的漏报率
            fail_study: True时返回失败案例的索引, 含漏报和误报的。False时返回空list。
        Return:
            min_score: 判断为ng的最小score, ng类别的阈值大于等于该score则判断为ng, 否则判断为ok 
            res: 含fnr和fpr
            fn_index_list: fail_study为False时为[], 否则为漏报的图片的index list
            fp_index_list: fail_study为False时为[], 否则为误报的图片的index list
        """
        y_true_new = []
        y_score_new = []
        for per_true, per_score in zip(y_true, y_score):
            label = 0 if per_true == ok_ind else 1
            per_score_new = per_score.copy()
            # 将ok类score置为0, 再取最大score, 该score为ng类别的最大score。
            per_score_new[ok_ind] = 0
            score = np.max(per_score_new)
            y_true_new.append(label)
            y_score_new.append(score)
        min_score, res, fn_index_list, fp_index_list = ClassifierEvalBinary.compute_fnr_and_fpr(
            np.array(y_true_new), np.array(y_score_new), fnr_thresh, fail_study)
        return min_score, res, fn_index_list, fp_index_list

    @classmethod
    def draw_failure_cases(cls, ok_ind, img_path_list, y_true, y_score, min_score, fn_index_list, fp_index_list, cls_dict, res_dir=None):
        """保存failure cases, 含漏报和误报的。
        Args:
            ok_ind: ok类别的index
            img_path_list: 图片路径
            y_true: 1维nparray, 所有样本的真值列表[nSamples], e.g., [2, 3, 1, 0]
            y_score: 2维nparray, 所有样本在各个类上的分数, e.g., [[0.2, 0.7, 0.1, 0], [0.1, 0.3, 0.5, 0.1], [0.4, 0.2, 0.1, 0.3], [0.3, 0.3, 0.4, 0]]
            min_score: 判为ng的最小score
            fn_index_list: 漏报的样本索引
            fp_index_list: 误报的样本索引
            cls_dict: 类别字典, 用于显示failure case, 如cls_dict={0:'ok', 1:'0', 2:'1', 3:'2', 4:'3', 5:'8'}
            res_dir: failure case存放路径
        Return:
            无
        """
        # Step 1. Make dirs.
        if not res_dir:
            raise ValueError('Result directory error!')
        res_dir_fn = os.path.join(res_dir, 'loubao')
        res_dir_fp = os.path.join(res_dir, 'wubao')
        if not os.path.exists(res_dir_fn):
            os.makedirs(res_dir_fn)
        if not os.path.exists(res_dir_fp):
            os.makedirs(res_dir_fp)

        # Step 2. Get dt_label_list.
        dt_label_list = []
        for per_score in y_score:
            per_score_new = per_score.copy()
            # 将ok类score置为0, 再取最大score, 该score为ng类别的最大score。
            per_score_new[ok_ind] = 0
            score = np.max(per_score_new)
            if score >= min_score:
                ind = np.argmax(per_score_new)
            else:
                ind = ok_ind
            dt_label_list.append(ind)

        # Step 3. Draw failures.
        if not fn_index_list:
            print('fn_index_list is empty!')
        else:
            for ind in fn_index_list:
                gt_label = y_true[ind]
                dt_label = dt_label_list[ind]
                img = mpimg.imread(img_path_list[ind])
                plt.imshow(img)
                plt.title(
                    'cls: {} -> cls: {}'.format(cls_dict[gt_label], cls_dict[dt_label]))
                plt.axis('off')
                output_path = os.path.join(
                    res_dir_fn, os.path.basename(img_path_list[ind]))
                output_path_new = output_path.replace('.jpg', '.png')
                plt.savefig(output_path_new)
                plt.close()

        if not fp_index_list:
            print('fp_index_list is empty!')
        else:
            for ind in fp_index_list:
                gt_label = y_true[ind]
                dt_label = dt_label_list[ind]
                img = mpimg.imread(img_path_list[ind])
                plt.imshow(img)
                plt.title(
                    'cls: {} -> cls: {}'.format(cls_dict[gt_label], cls_dict[dt_label]))
                plt.axis('off')
                output_path = os.path.join(
                    res_dir_fp, os.path.basename(img_path_list[ind]))
                output_path_new = output_path.replace('.jpg', '.png')
                plt.savefig(output_path_new)
                plt.close()


class ClassifierEvalMultilabel(ClassifierEval):
    """多标签分类评估器。"""

    @classmethod
    def compute_ap(cls, y_true, y_score, class_name_list = None):
        """计算多标签分类的AP 
        Args:
            y_true: 2维nparray, 所有样本的真值列表[nSamples, nClasses], e.g., [[0, 0, 1, 0], [1, 0, 0, 0]]
            y_score: 2维nparray, 所有样本在各个类别上的分数, 注意每一个样本在各个类别上的分数之和须为1, e.g., [[0.2, 0.3, 0.4, 0.1], [0.1, 0.1, 0.3, 0.5], ...]
            class_name_list: str list, 各个类别名称列表 ['plane', 'hot_dog', 'river', 'girl']
        Return:
            ap_dict: 各个类别的AP值组成的dict, key值为序号(class_name_list == None)/key值为类别名称(class_name_list != None)
        """
        ap_dict = dict()
        assert y_true.shape == y_score.shape and np.ndim(y_true) == 2
        if class_name_list:
            assert isinstance(class_name_list, list)
            assert len(class_name_list) == y_true.shape[1]
        for i, (true_per_cls, score_per_cls) in enumerate(zip(y_true.T, y_score.T)):
            ap = ClassifierEvalBinary.compute_ap(y_true=true_per_cls, y_score=score_per_cls)
            if class_name_list:
                key = class_name_list[i]
            else:
                key = i
            if np.isnan(ap):
                ap = 0
            ap_dict[key] = ap

        return ap_dict

    @classmethod
    def compute_p_at_r(cls, y_true, y_score, class_name_list=None, recall_thresh=0.995):
        """计算多标签中recall >= 0.995时各个类别的precision。
        Args:
            y_true: 2维nparray, 所有样本的真值列表[nSamples, nClasses], e.g., [[0, 0, 1, 0], [1, 0, 0, 0], [0, 0, 0, 1]]
            y_score: 1维nparray, 所有样本在各个类上的分数, e.g., [[0.2, 0.7, 0.1, 0], [0.1, 0.3, 0.5, 0.1], [0.4, 0.2, 0.1, 0.3]]
        Return:
            每个类别的Precision@(Recall>=recall_thresh)组成的dict, 如{0: 0.2345, 1: 0.2344, 2: 0.7623, 3: 0.3334}
        """
        prec_dict = dict()
        for i, (true_per_cls, score_per_cls) in enumerate(zip(y_true.T, y_score.T)):
            precision = ClassifierEvalBinary.compute_p_at_r(
                true_per_cls, score_per_cls, recall_thresh=recall_thresh)
            if class_name_list:
                key = class_name_list[i]
            else:
                key = i
            prec_dict[key] = round(precision, 4)  # 保留4位小数
        return prec_dict

    @classmethod
    def draw_pr_curve(cls, y_true, y_score, output_dir='hehe', class_name_list=None):
        """画多标签分类的pr曲线, 每个类别一个pr曲线。
        Args:
            y_true: 2维nparray, 所有样本的真值列表[nSamples, nClasses], e.g., [[0, 0, 1, 0], [1, 0, 0, 0], [0, 0, 0, 1]]
            y_score: 1维nparray, 所有样本在各个类上的分数, e.g., [[0.2, 0.7, 0.1, 0], [0.1, 0.3, 0.5, 0.1], [0.4, 0.2, 0.1, 0.3]]
            class_name_list: str list, 各个类别名称列表 ['plane', 'hot_dog', 'river', 'girl']
        Return:
            无
        """
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        assert y_true.shape == y_score.shape and np.ndim(y_true) == 2
        if class_name_list:
            assert isinstance(class_name_list, list)
            assert len(class_name_list) == y_true.shape[1]

        for i, (true_per_cls, score_per_cls) in enumerate(zip(y_true.T, y_score.T)):
            if class_name_list:
                class_name = class_name_list[i]
            else:
                class_name = 'cls_' + str(i)
            output_path = os.path.join(output_dir, class_name)
            ClassifierEvalBinary.draw_pr_curve(true_per_cls, score_per_cls, output_path=output_path, class_name=class_name)

    @classmethod
    def get_inflection_info(cls, y_true, y_score, class_name_list=None):
        """
        计算拐点信息与ap信息
        args:
            y_true: 2维nparray, 所有样本的真值列表[nSamples, nClasses], e.g., [[0, 0, 1, 0], [1, 0, 0, 0], [0, 0, 0, 1]]
            y_score: 1维nparray, 所有样本在各个类上的分数, e.g., [[0.2, 0.7, 0.1, 0], [0.1, 0.3, 0.5, 0.1], [0.4, 0.2, 0.1, 0.3]]
        returns:
            class_ap_info_dict: 包含拐点与AP信息的字典
        """
        assert y_true.shape == y_score.shape and np.ndim(y_true) == 2
        if class_name_list:
            assert isinstance(class_name_list, list)
            assert len(class_name_list) == y_true.shape[1]
        class_ap_info_dict = dict()
        for i, (true_per_cls, score_per_cls) in enumerate(zip(y_true.T, y_score.T)):
            if class_name_list:
                class_name = class_name_list[i]
            else:
                class_name = 'cls_' + str(i)
            inflection_info_dict = ClassifierEvalBinary.get_inflection_info(true_per_cls, score_per_cls)
            class_ap_info_dict[class_name] = inflection_info_dict

        return class_ap_info_dict

    @classmethod
    def compute_fnr_and_fpr(cls, y_true, y_score, ok_ind=0, fnr_thresh=0.005, fail_study=False):
        """计算多标签分类中的漏报率和误报率, 要求漏报率必须小于等于fnr_thresh。
        Args:
            y_true: 2维nparray, 所有样本的真值列表[nSamples, nClasses], e.g., [[0, 0, 1, 0], [0, 0, 0, 1], [0, 1, 0, 0], [1, 0, 0, 0]]
            y_score: 2维nparray, 所有样本在各个类上的分数, e.g., [[0.2, 0.7, 0.1, 0], [0.1, 0.3, 0.5, 0.1], [0.4, 0.2, 0.1, 0.3], [0.3, 0.3, 0.4, 0]]
            ok_ind: ok类别的index, 默认0为ok, 其余为ng 
            fnr_thresh: 能容忍的最大的漏报率
            fail_study: True时返回失败案例的索引, 含漏报和误报的。False时返回空list。
        Return:
            min_score: 判断为ng的最小score, ng类别的阈值大于等于该score则判断为ng, 否则判断为ok 
            res: 含fnr和fpr
            fn_index_list: fail_study为False时为[], 否则为漏报的图片的index list
            fp_index_list: fail_study为False时为[], 否则为误报的图片的index list
        """
        y_true_new = []
        y_score_new = []
        for per_true, per_score in zip(y_true, y_score):
            per_true_array = np.where(per_true == 1)[0]
            label = 0 if ok_ind in per_true_array else 1
            per_score_new = per_score.copy()
            # 将ok类score置为0, 再取最大score, 该score为ng类别的最大score。
            per_score_new[ok_ind] = 0
            score = np.max(per_score_new)
            y_true_new.append(label)
            y_score_new.append(score)
        min_score, res, fn_index_list, fp_index_list = ClassifierEvalBinary.compute_fnr_and_fpr(
            np.array(y_true_new), np.array(y_score_new), fnr_thresh, fail_study)
        return min_score, res, fn_index_list, fp_index_list

    @classmethod
    def draw_failure_cases(cls, ok_ind, img_path_list, y_true, y_score, min_score, fn_index_list, fp_index_list, cls_dict, res_dir=None):
        """保存failure cases, 含漏报和误报的。
        Args:
            ok_ind: ok类别的index
            img_path_list: 图片路径
            y_true: 2维nparray, 所有样本的真值列表[nSamples, nClasses], e.g., [[0, 0, 1, 0], [0, 0, 0, 1], [0, 1, 0, 0], [1, 0, 0, 0]]
            y_score: 2维nparray, 所有样本在各个类上的分数, e.g., [[0.2, 0.7, 0.1, 0], [0.1, 0.3, 0.5, 0.1], [0.4, 0.2, 0.1, 0.3], [0.3, 0.3, 0.4, 0]]
            min_score: 判为ng的最小score
            fn_index_list: 漏报的样本索引
            fp_index_list: 误报的样本索引
            cls_dict: 类别字典, 用于显示failure case, 如cls_dict={0:'ok', 1:'0', 2:'1', 3:'2', 4:'3', 5:'8'}
            res_dir: failure case存放路径
        Return:
            无
        """
        # Step 1. Make dirs.
        if not res_dir:
            raise ValueError('Result directory error!')
        res_dir_fn = os.path.join(res_dir, 'loubao')
        res_dir_fp = os.path.join(res_dir, 'wubao')
        if not os.path.exists(res_dir_fn):
            os.makedirs(res_dir_fn)
        if not os.path.exists(res_dir_fp):
            os.makedirs(res_dir_fp)

        # Step 2. Get dt_label_list.
        gt_label_list = []
        dt_label_list = []
        for per_true, per_score in zip(y_true, y_score):
            per_true_array = np.where(per_true == 1)[0]
            gt_label_list.append(list(per_true_array))
            per_score_new = per_score.copy()
            # 将ok类score置为0, 再取最大score, 该score为ng类别的最大score。
            per_score_new[ok_ind] = 0
            score = np.max(per_score_new)
            if score >= min_score:
                ind = list(np.where(per_score_new >= min_score)[0])
            else:
                ind = [ok_ind]
            dt_label_list.append(ind)

        # Step 3. Draw failures.
        if not fn_index_list:
            print('fn_index_list is empty!')
        else:
            for ind in fn_index_list:
                gt_label = gt_label_list[ind]
                dt_label = dt_label_list[ind]
                gt_label_show = [cls_dict[lab] for lab in gt_label]
                dt_label_show = [cls_dict[lab] for lab in dt_label]
                img = mpimg.imread(img_path_list[ind])
                plt.imshow(img)
                plt.title(
                    'cls: {} -> cls: {}'.format(gt_label_show, dt_label_show))
                plt.axis('off')
                output_path = os.path.join(
                    res_dir_fn, os.path.basename(img_path_list[ind]))
                output_path_new = output_path.replace('.jpg', '.png')
                plt.savefig(output_path_new)
                plt.close()

        if not fp_index_list:
            print('fp_index_list is empty!')
        else:
            for ind in fp_index_list:
                gt_label = gt_label_list[ind]
                dt_label = dt_label_list[ind]
                gt_label_show = [cls_dict[lab] for lab in gt_label]
                dt_label_show = [cls_dict[lab] for lab in dt_label]
                img = mpimg.imread(img_path_list[ind])
                plt.imshow(img)
                plt.title(
                    'cls: {} -> cls: {}'.format(gt_label_show, dt_label_show))
                plt.axis('off')
                output_path = os.path.join(
                    res_dir_fp, os.path.basename(img_path_list[ind]))
                output_path_new = output_path.replace('.jpg', '.png')
                plt.savefig(output_path_new)
                plt.close()

    @classmethod
    def get_threshold_at_p(cls, y_true, y_score, precision_thresh=0.9, class_name_list=None):
        """
        计算各个类别在precision处于precision_thresh水平下最小阈值, 不存在则返回最大int值
        args:
            y_true: 2维nparray, 所有样本的真值列表[nSamples, nClasses], e.g., [[0, 0, 1, 0], [0, 0, 0, 1], [0, 1, 0, 0], [1, 0, 0, 0]]
            y_score: 2维nparray, 所有样本在各个类上的分数, e.g., [[0.2, 0.7, 0.1, 0], [0.1, 0.3, 0.5, 0.1], [0.4, 0.2, 0.1, 0.3], [0.3, 0.3, 0.4, 0]]
            precision_thresh: precision 阈值 [0, 1]
            class_name_list: str list, 各个类别名称列表 ['plane', 'hot_dog', 'river', 'girl']
        return:
            thre_dict: 各个类别的阈值组成的dict, key值为序号(class_name_list == None)/key值为类别名称(class_name_list != None)
        """
        assert 0 <= precision_thresh <= 1, " the <precision_thresh> should be a float in [0, 1], while the input is {}".format(precision_thresh)
        thre_dict = dict()
        for i, (true_per_cls, score_per_cls) in enumerate(zip(y_true.T, y_score.T)):
            thre = ClassifierEvalBinary.get_threshold_at_p(true_per_cls, score_per_cls, precision_thresh)
            if class_name_list:
                thre_dict[class_name_list[i]] = thre
            else:
                thre_dict[i] = thre
        return thre_dict

if __name__ == '__main__':
    pass



import copy
import numpy as np

from .format import DataManagerFormater
from ..utils import decode_distribution
from ..utils import decode_labelme_shape


class DataManagerInfoExtractor(DataManagerFormater):
    @property
    def task_type(self):
        record_types = set(self.extract_info(self._get_record_task_type))
        assert len(record_types) in [0, 1], "Multiple task types exist in one dataset!"
        return record_types.pop() if len(record_types) else 'detection'

    def class2label(self, classname):
        for classid, name in enumerate(self.class_list):
            if name == classname:
                return classid
        raise RuntimeError("Unrecognized class: {}".format(classname))

    def label2class(self, classid):
        return self.class_list[classid]

    ## Implementation of different classname getting stuffs
    # 1) get classname
    def get_detection_classname_list_from_record(self, record):
        assert self._get_record_task_type(record) == 'detection'
        return [self.get_detection_classname_from_instance(inst) for inst in record['instances']]

    def get_detection_classname_from_instance(self, instance):
        return instance['classname']

    def get_multiclass_classname_from_record(self, record):
        assert self._get_record_task_type(record) == 'multiclass'
        distribution = decode_distribution(record['distribution'])
        return self.get_multiclass_classname_from_distribution(distribution)

    def get_multiclass_classname_from_distribution(self, distribution):
        assert len(distribution) == len(self.class_list)
        classid = np.argmax(distribution)
        return self.label2class(classid)

    def get_classname_from_classid(self, classid):
        return self.label2class(classid)

    # 2) get classid
    def get_detection_classid_list_from_record(self, record):
        assert self._get_record_task_type(record) == 'detection'
        return [self.get_detection_classid_from_instance(inst) for inst in record['instances']]

    def get_detection_classid_from_instance(self, instance):
        return self.class2label(instance['classname'])

    def get_multiclass_classid_from_record(self, record):
        assert self._get_record_task_type(record) == 'multiclass'
        distribution = decode_distribution(record['distribution'])
        return self.get_multiclass_classid_from_distribution(distribution)

    def get_multiclass_classid_from_distribution(self, distribution):
        assert len(distribution) == len(self.class_list)
        classid = np.argmax(distribution)
        return classid

    def get_classid_from_classname(self, classname):
        return self.class2label(classname)

    # 3) get score (detection/multiclass)
    def get_detection_score_list_from_record(self, record):
        assert self._get_record_task_type(record) == 'detection'
        return [self.get_detection_score_from_instance(inst) for inst in record['instances']]
    
    def get_detection_score_from_instance(self, instance):
        return instance['score']

    def get_multiclass_score_from_record(self, record):
        distribution = self.get_multiclass_distribution_from_record(record)
        return self.get_multiclass_score_from_distribution(distribution)

    def get_multiclass_score_from_distribution(self, distribution):
        distribution = decode_distribution(distribution)
        return np.max(distribution)

    # 4) get scores (multilabel)
    def get_multilabel_scores_from_record(self, record):
        assert self._get_record_task_type(record) == 'multilabel'
        return decode_distribution(record['scores'])

    # 5) get distribution (multiclass)
    def get_multiclass_distribution_from_record(self, record):
        assert self._get_record_task_type(record) == 'multiclass'
        return decode_distribution(record['distribution'])

    # 6) get shape/xyxy/xywh (detection)
    def get_detection_shape_list_from_record(self, record):
        assert self._get_record_task_type(record) == 'detection'
        return [self.get_detection_shape_from_instance(inst) for inst in record['instances']]

    def get_detection_shape_from_instance(self, instance):
        assert 'shape_type' in instance
        if None: pass
        elif instance['shape_type'] == 'polygon':
            points = decode_labelme_shape(instance['points'])
            points = np.reshape(points, [-1,2])
            return points.tolist()
        elif instance['shape_type'] == 'rectangle':
            x1,y1,x2,y2 = decode_labelme_shape(instance['points'])
            return [[x1,y1],[x1,y2],[x2,y2],[x2,y1],[x1,y1]]
        else:
            raise RuntimeError("Unknown shape type: {}".format(instance['shape_type']))

    def get_detection_xyxy_list_from_record(self, record):
        assert self._get_record_task_type(record) == 'detection'
        return [self.get_detection_xyxy_from_instance(inst) for inst in record['instances']]

    def get_detection_xyxy_from_instance(self, instance):
        points = self.get_detection_shape_from_instance(instance)
        points = np.reshape(points, [-1,2])
        x1,y1,x2,y2 = points[:,0].min(), points[:,1].min(), points[:,0].max(), points[:,1].max()
        return [x1,y1,x2,y2]

    def get_detection_xywh_list_from_record(self, record):
        assert self._get_record_task_type(record) == 'detection'
        return [self.get_detection_xywh_from_instance(inst) for inst in record['instances']]

    def get_detection_xywh_from_instance(self, instance):
        x1,y1,x2,y2 = self.get_detection_xyxy_from_instance(instance)
        return [x1, y1, x2-x1, y2-y1]

    def data_statistics(self):
        data_task_type = self.task_type
        if data_task_type == 'detection':
            get_classname = self.get_detection_classname_list_from_record
            class_occurence_by_instance = self.occurrence(get_classname)
            class_occurence_by_image = self.occurrence(lambda rec: list(set(get_classname(rec))))
            ok_ng_counter = self.occurrence(lambda rec: len(rec['instances'])>0)
            return {
                '#images': len(self),
                '#ok_images': ok_ng_counter.get(False, 0),
                '#ng_images': ok_ng_counter.get(True,  0),
                '#classes_by_image': class_occurence_by_image,
                '#classes_by_instance': class_occurence_by_instance
            }
        elif data_task_type == 'multiclass':
            class_occurrence = self.occurrence(self.get_multiclass_classname_from_record)
            return {
                '#images': len(self),
                '#classes': class_occurrence
            }
        else:
            assert data_task_type == 'multilabel'
            return {'#images': len(self)}